import { describe, it, expect, beforeEach } from 'vitest'
import * as rollup from 'rollup'
import { importFromString } from 'module-from-string'

import gettextPlugin from '../dist/index'
import { itemKey } from '../dist/util/util'

describe('Dictionary API', () => {
  let dictionary

  beforeEach(async () => {
    const { output } = await render()
    const module = output.find(({ fileName }) => fileName === 'test/mySimpleModule.js')
    ;({ dictionary } = await importFromString(module.code))
  })

  async function render (opts = {}) {
    const bundle = await rollup.rollup({
      input: {
        'test/mySimpleModule': 'gettext'
      },
      plugins: [gettextPlugin(Object.assign({
        poFiles: 'spec/fixtures/multiple_languages/*.po',
        defaultDictionary: 'test/mySimpleModule',
        defaultLanguage: 'de_DE'
      }, opts))]
    })
    return await bundle.generate({})
  }

  function addTranslation (message, translation) {
    dictionary.dict[itemKey(message)] = translation
  }

  function removeTranslation (message) {
    delete dictionary.dict[itemKey(message)]
  }

  describe('simple translations', () => {
    it('should simply translate the string', () => {
      const message = { msgid: 'Foobar' }
      addTranslation(message, ['I am translated'])
      expect(dictionary.gettext('Foobar')).toBe('I am translated')
      removeTranslation(message)
      expect(dictionary.gettext('Foobar')).toBe('Foobar')
    })
  })

  describe('plural forms', () => {
    const message = { msgid: 'Foobar', msgid_plural: 'Foobars' }
    it('should translate strings using ngettext function', () => {
      addTranslation(message, ['I am translated', 'Plural'])
      expect(dictionary.ngettext('Foobar', 'Foobars', 1)).toBe('I am translated')
      expect(dictionary.ngettext('Foobar', 'Foobars', 2)).toBe('Plural')
      removeTranslation(message)
      expect(dictionary.ngettext('Foobar', 'Foobars', 1)).toBe('Foobar')
      expect(dictionary.ngettext('Foobar', 'Foobars', 2)).toBe('Foobars')
    })

    it('should support more than 2 plural translations', () => {
      addTranslation(message, ['I am translated', 'Plural 1', 'Plural 2'])
      const oldHeaders = Object.assign({}, dictionary.headers)
      dictionary.headers = {
        'Plural-Forms': 'nplurals=2; plural=n;\n'
      }
      expect(dictionary.ngettext('Foobar', 'Foobars', 0)).toBe('I am translated')
      expect(dictionary.ngettext('Foobar', 'Foobars', 1)).toBe('Plural 1')
      expect(dictionary.ngettext('Foobar', 'Foobars', 2)).toBe('Plural 2')
      // fallback to singular form
      expect(dictionary.ngettext('Foobar', 'Foobars', 3)).toBe('I am translated')
      dictionary.headers = oldHeaders
    })

    it('should provide defaults for plural form implementation', () => {
      addTranslation(message, ['I am translated', 'Plural'])
      const oldHeaders = Object.assign({}, dictionary.headers)
      dictionary.headers = {}

      expect(dictionary.ngettext('Foobar', 'Foobars', 1)).toBe('I am translated')
      expect(dictionary.ngettext('Foobar', 'Foobars', 2)).toBe('Plural')

      dictionary.headers = oldHeaders
    })
  })

  describe('message contexts', () => {
    const message = { msgid: 'Foobar', msgctxt: 'noun' }
    it('should translate string using pgettext function', () => {
      addTranslation(message, ['I am translated'])
      expect(dictionary.pgettext('noun', 'Foobar')).toBe('I am translated')
      removeTranslation(message)
      expect(dictionary.pgettext('noun', 'Foobar')).toBe('Foobar')
    })
  })

  describe('plural forms with message contexts', () => {
    const message = { msgctxt: 'noun', msgid: 'Foobar', msgid_plural: 'Foobars' }
    it('should translate strings using npgettext function', () => {
      addTranslation(message, ['I am translated', 'Plural'])
      expect(dictionary.npgettext('noun', 'Foobar', 'Foobars', 1)).toBe('I am translated')
      expect(dictionary.npgettext('noun', 'Foobar', 'Foobars', 2)).toBe('Plural')
      removeTranslation(message)
      expect(dictionary.npgettext('noun', 'Foobar', 'Foobars', 1)).toBe('Foobar')
      expect(dictionary.npgettext('noun', 'Foobar', 'Foobars', 2)).toBe('Foobars')
    })
  })

  describe('implicit c-string formatting', () => {
    it('should print string parameters', () => {
      expect(dictionary.gettext('Foo %1$s', 'bar')).toBe('Foo bar')
    })
    it('should print number parameters', () => {
      expect(dictionary.gettext('Foo %1$d', 100)).toBe('Foo 100')
    })
  })

  describe('change language', () => {
    it('should change the language', async () => {
      await expect(async () => await dictionary.changeLanguage('en_US')).rejects.toThrow('Cannot find module')
    })

    it('should ignore language changes with unknown language', async () => {
      await expect(async () => await dictionary.changeLanguage('non-existing')).rejects.toThrow('Language not found')
    })
  })
})
