import { describe, it, expect } from 'vitest'
import { build, createServer } from 'vite'
import vitePluginOxExternals from '../../src/index'
import { getFileFromBundle, getPort } from '../util'

const __dirname = import.meta.dirname

describe('Simple scenario', () => {
  it('works', async () => {
    const bundle = await build({
      root: __dirname,
      logLevel: 'silent',
      build: {
        write: false,
        rollupOptions: {
          output: {
            entryFileNames: 'entries/[name].js'
          }
        }
      },
      plugins: [vitePluginOxExternals()]
    })

    const indexJS = getFileFromBundle('index.js', bundle)
    expect(indexJS.code).toContain('"../module/external.js"')
    expect(indexJS.code).toContain('"../external.js"')
    expect(indexJS.code).toContain('No replace: import mod from "$/module/external"')
    expect(indexJS.imports).toEqual(['module/external.js', 'external.js'])
  })

  it('works with minified disabled', async () => {
    const bundle = await build({
      root: __dirname,
      logLevel: 'silent',
      build: {
        write: false,
        minify: false,
        rollupOptions: {
          output: {
            entryFileNames: 'entries/[name].js'
          }
        }
      },
      plugins: [vitePluginOxExternals({
        prefix: '$'
      })]
    })

    const indexJS = getFileFromBundle('index.js', bundle)
    expect(indexJS.code).toContain('"../module/external.js"')
    expect(indexJS.code).toContain('"../external.js"')
    expect(indexJS.code).toContain('No replace: import mod from "$/module/external"')
    expect(indexJS.imports).toEqual(['module/external.js', 'external.js'])
  })

  it('works in dev mode', async () => {
    const server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginOxExternals({
        prefix: '$'
      })]
    })
    const PORT = await getPort()
    await server.listen(PORT)

    const res = await fetch(`http://localhost:${PORT}/index.js`)
    expect(await res.text()).toMatch('import mod from "/module/external.js"\nimport "/external.js"\n\nconsole.log(\'No replace: import mod from "$/module/external"\', mod)\n')

    await server.close()
  })

  it('works in dev mode with different base', async () => {
    const server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/appsuite/',
      plugins: [vitePluginOxExternals({
        prefix: '$'
      })]
    })
    const PORT = await getPort()
    await server.listen(PORT)

    const res = await fetch(`http://localhost:${PORT}/appsuite/index.js`)
    expect(await res.text()).toMatch('import mod from "/appsuite/module/external.js"\nimport "/appsuite/external.js"\n\nconsole.log(\'No replace: import mod from "$/module/external"\', mod)\n')

    await server.close()
  })
})
