import path from 'node:path'
import { readFileSync, writeFileSync } from 'node:fs'
import { describe, it, expect, beforeEach, afterEach } from 'vitest'
import { createServer } from 'vite'
import vitePluginOxExternals from '@open-xchange/vite-plugin-ox-externals'

import vitePluginOxManifests from '../../dist/index'
import { waitForReload } from '../../dist/util'
import { getPort } from '../util'

const __dirname = import.meta.dirname
const files = ['register.js', 'settings.js']

describe('HMR', () => {
  let data

  beforeEach(() => {
    data = files.map(f => readFileSync(path.join(__dirname, f)))
  })

  afterEach(async () => {
    data.forEach((d, i) => writeFileSync(path.join(__dirname, files[i]), d))
    data = null
  })

  it('adds timestamps to manifest files', async () => {
    const server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [
        vitePluginOxManifests({ }),
        vitePluginOxExternals({ prefix: '$' })
      ]
    })
    const PORT = await getPort()
    await server.listen(PORT)

    // fetch this file to start watcher
    await fetch(`http://localhost:${PORT}/register.js`)

    const manifests = await fetch(`http://localhost:${PORT}/manifests`).then(r => r.json())
    expect(Object.keys(manifests)).toHaveLength(2)
    expect(manifests[1]).not.toContain('raw')

    writeFileSync(path.join(__dirname, 'register.js'), 'console.log("Changed")')

    await waitForReload(server)

    const manifestsWithHMR = await fetch(`http://localhost:${PORT}/manifests`).then(r => r.json())
    expect(Object.keys(manifestsWithHMR)).toHaveLength(2)
    expect(manifestsWithHMR[1].namespace).toBe('test')
    expect(manifestsWithHMR[1].raw).toMatch(/\/register\.js\?t=\d+/)

    await server.close()
  })

  it('adds timestamps to settings manifest files', async () => {
    const server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginOxManifests({

      }), vitePluginOxExternals({
        prefix: '$'
      })]
    })
    const PORT = await getPort()
    await server.listen(PORT)

    // fetch this file to start watcher
    await fetch(`http://localhost:${PORT}/settings.js`)

    const manifests = await fetch(`http://localhost:${PORT}/manifests`).then(r => r.json())
    expect(Object.keys(manifests)).toHaveLength(2)
    expect(manifests[0]).not.toContain('raw')

    writeFileSync(path.join(__dirname, 'settings.js'), `/* a comment */${data[1].toString()}`)

    await waitForReload(server)

    const manifestsWithHMR = await fetch(`http://localhost:${PORT}/manifests`).then(r => r.json())
    expect(Object.keys(manifestsWithHMR)).toHaveLength(2)
    expect(manifestsWithHMR[0].namespace).toBe('settings')
    expect(manifestsWithHMR[0].raw).toMatch(/\/settings\.js\?t=\d+/)
    await server.close()
  })

  it('includes base-path in manifests', async () => {
    const server = await createServer({
      root: __dirname,
      logLevel: 'silent',
      base: '/appsuite',
      plugins: [vitePluginOxManifests({

      }), vitePluginOxExternals({
        prefix: '$'
      })]
    })
    const PORT = await getPort()
    await server.listen(PORT)

    // fetch this file to start watcher
    await fetch(`http://localhost:${PORT}/appsuite/register.js`)
    await fetch(`http://localhost:${PORT}/appsuite/settings.js`)

    const manifests = await fetch(`http://localhost:${PORT}/appsuite/manifests`).then(r => r.json())
    expect(Object.keys(manifests)).toHaveLength(2)
    expect(manifests[0]).not.toContain('raw')
    expect(manifests[1]).not.toContain('raw')

    writeFileSync(path.join(__dirname, 'register.js'), 'console.log("Changed")')
    await waitForReload(server)

    writeFileSync(path.join(__dirname, 'settings.js'), `/* a comment */${data[1].toString()}`)
    await waitForReload(server)

    const manifestsWithHMR = await fetch(`http://localhost:${PORT}/appsuite/manifests`).then(r => r.json())
    expect(Object.keys(manifestsWithHMR)).toHaveLength(2)
    expect(manifestsWithHMR[0].namespace).toBe('settings')
    expect(manifestsWithHMR[0].raw).toMatch(/\/appsuite\/settings\.js\?t=\d+/)
    expect(manifestsWithHMR[1].namespace).toBe('test')
    expect(manifestsWithHMR[1].raw).toMatch(/\/appsuite\/register\.js\?t=\d+/)
    await server.close()
  })
})
