import { createServer } from 'vite'
import { describe, it, expect, afterEach } from 'vitest'
import vitePluginProxy from '../../dist/index'
import vitePluginOxManifests from '@open-xchange/vite-plugin-ox-manifests'
import { createTestServer, getPort, http2Request, updateRequestManager } from '../util'
import fastifyCompress from '@fastify/compress'

const __dirname = import.meta.dirname
process.noDeprecation = true

describe('manifests', () => {
  afterEach(async () => {
    updateRequestManager()
  })

  it('serves merged manifest.json', async () => {
    const server = await createTestServer()
    server.get('/manifests', (req, res) => {
      res.send([{ namespace: 'other' }])
    })
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        frontends: [{ target: `http://localhost:${SERVER_PORT}` }]
      }), vitePluginOxManifests({
        // no config
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/manifests`)
    expect(data).toEqual([{
      namespace: 'other'
    }, {
      namespace: 'index',
      path: 'index'
    }])
    await server1.close()
    await server.close()
  })

  it('overrides manifest.json with local manifests', async () => {
    const server = await createTestServer()
    server.get('/manifests', (req, res) => {
      res.send([{ namespace: 'index', path: 'index', requires: 'chat' }])
    })
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        // this has to be thought of as a remote UI with the same files
        frontends: [{ target: `http://localhost:${SERVER_PORT}` }]
      }), vitePluginOxManifests({
        // no config
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/manifests`)
    expect(data).toEqual([{
      namespace: 'index',
      path: 'index'
    }])
    await server1.close()
    await server.close()
  })

  describe('compression', () => {
    it('serves gzipped manifests', async () => {
      const server = await createTestServer()
      server.register(fastifyCompress, { threshold: 1, forceRequestEncoding: 'gzip', encodings: ['gzip'] })
      server.get('/manifests', (req, res) => {
        res.send([{ namespace: 'other' }])
      })
      const SERVER_PORT = await getPort()
      await server.listen({ port: SERVER_PORT })

      const server1 = await createServer({
        root: __dirname,
        logLevel: 'silent',
        plugins: [vitePluginProxy({
          frontends: [{ target: `http://localhost:${SERVER_PORT}` }]
        }), vitePluginOxManifests({
          // no config
        })]
      })
      const PORT = await getPort()
      await server1.listen(PORT)

      const { data } = await http2Request(`http://localhost:${PORT}/manifests`)

      expect(data).toEqual([{
        namespace: 'other'
      }, {
        namespace: 'index',
        path: 'index'
      }])

      await server1.close()
      await server.close()
    })

    it('serves inflated manifests', async () => {
      const server = await createTestServer()
      server.register(fastifyCompress, { threshold: 1, forceRequestEncoding: 'deflate', encodings: ['deflate'] })
      server.get('/manifests', (req, res) => {
        res.send([{ namespace: 'other' }])
      })
      const SERVER_PORT = await getPort()
      await server.listen({ port: SERVER_PORT })

      const server1 = await createServer({
        root: __dirname,
        logLevel: 'silent',
        plugins: [vitePluginProxy({
          frontends: [{ target: `http://localhost:${SERVER_PORT}` }]
        }), vitePluginOxManifests({
          // no config
        })]
      })
      const PORT = await getPort()
      await server1.listen(PORT)

      const { data } = await http2Request(`http://localhost:${PORT}/manifests`)

      expect(data).toEqual([{
        namespace: 'other'
      }, {
        namespace: 'index',
        path: 'index'
      }])
    })
  })

  it('serves merged manifest.json from a different base', async () => {
    const server = await createTestServer()
    server.get('/appsuite/manifests', (req, res) => {
      res.send([{ namespace: 'other' }])
    })
    const SERVER_PORT = await getPort()
    await server.listen({ port: SERVER_PORT })

    const server1 = await createServer({
      root: __dirname,
      base: '/appsuite/',
      logLevel: 'silent',
      plugins: [vitePluginProxy({
        frontends: [{ target: `http://localhost:${SERVER_PORT}` }]
      }), vitePluginOxManifests({
        // no config
      })]
    })
    const PORT = await getPort()
    await server1.listen(PORT)

    const { data } = await http2Request(`http://localhost:${PORT}/appsuite/manifests`)
    expect(data).toEqual([{
      namespace: 'other'
    }, {
      namespace: 'index',
      path: 'index'
    }])
    await server1.close()
    await server.close()
  })
})
